#!/usr/bin/env python3
#
# \u7ADC TatSu documentation build configuration file, created by
# sphinx-quickstart on Mon May  1 18:01:31 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('.'))
del sys
del os
import tatsu

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',

    "myst_parser",
    "sphinx_copybutton",  # Recommended for "Copy" icons on code blocks
]

# MyST-Parser Settings
myst_enable_extensions = [
    "colon_fence",     # Allows using ::: for notes/admonitions
    "deflist",         # Allows definition lists (Term : Definition)
    "dollarmath",      # Allows LaTeX math via $$ ... $$
    "fieldlist",       # Allows GitHub-style metadata fields
    "tasklist",        # Allows - [ ] checkbox lists
    "html_image",      # Allows using standard <img src="..."> tags
]

# Auto-generate anchors for headers (so you can link to #my-header)
myst_heading_anchors = 4  # Generates anchors for H1, H2, H3, ...

copybutton_prompt_text = r">>> |\$ "
copybutton_only_copy_prompt_lines = True
copybutton_use_regexp = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
# source_suffix = '.rst'
source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = '\u7ADC TatSu'
copyright = '2017-2026 Juancarlo Añez'
author = 'Juancarlo Añez'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = tatsu.__version__
# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = 'sphinx_rtd_theme'
# html_theme = 'alabaster'
html_theme = "pydata_sphinx_theme"

html_context = {
   "default_mode": "auto"  # Options: "light", "dark", or "auto"
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}
html_theme_options = {
    # Controls how deep the sidebar navigation goes
    'navigation_depth': 5,

    # If True, the sidebar will only show document titles, not subheadings
    # 'titles_only': False,

    # If True, headings will collapse as you navigate
    'collapse_navigation': True,

    "logo": {
        "text": "TatSu",
        "image_light": "_static/tatsu_logo.svg",
        "image_dark": "_static/tatsu_logo.svg",
    },
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/neogeny/TatSu",
            "icon": "fa-brands fa-github",
        },
    ],
    "header_links_before_dropdown": 4,  # How many nav links to show before "More"
    "show_prev_next": True,             # Previous/Next buttons at the bottom
    "navbar_align": "content",          # Align the top nav with the text content
}

# Add any paths that contain custom themes here, relative to this directory.
# Add path to the RTD explicitly to robustify builds (otherwise might
# fail in a clean Debian build env)
html_theme_path = []

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static', '_static/css']


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'tatsudoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'tatsu.tex', '\u7ADC TatSu Documentation',
     'Juancarlo Añez', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, '\u7ADC TatSu', '\u7ADC TatSu Documentation',
     [author], 1),
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, '\u7ADC TatSu', '\u7ADC TatSu Documentation',
     author, '\u7ADC TatSu', 'One line description of project.',
     'Miscellaneous'),
]


html_sidebars = {
    '**': [
        'globaltoc.html',
        'relations.html',
        'sourcelink.html',
        'searchbox.html',
    ],
}


def setup(app):
    app.add_css_file('css/custom.css')  # may also be an URL
